#include "../Iinstr.hpp"

class Ccmpi : public Iinstr {
	public:
		Ccmpi();
		void evalInstr();
		void execInstr();
		std::string toString();
		Iinstr* getNew();
};

/**
 * Name:      create_instruction
 * Purpose:   if compiled as shared library, this functions creates the 
              instruction object

 * Returns:   pointer to instruction object
 */
extern "C" Iinstr* create_instruction() {
    return new Ccmpi();
}

Iinstr* Ccmpi::getNew() {
	return new Ccmpi();
}

/**
 * Name:      destroy_instruction
 * Purpose:   if compiled as shared library, this functions destoys the 
              instruction object

 * Parameter: IInstruction - the instruction object to delete
 */
extern "C" void destroy_instruction(Iinstr* p) {
    delete p;
}

Ccmpi::Ccmpi()
{
	opcode = B5(11001);
	name = "cmpi";
}


void Ccmpi::evalInstr()
{
	argbits >>= 3;
	dynamic_bitset<> immb = argbits;
	immb.resize(16);
	this->m_imm = this->generate16ImmSign(immb.to_ulong());

	argbits >>= 16;
	m_ra = this->getRegister(argbits);
}

void Ccmpi::execInstr()
{
	//cout << "should exec " << this->toString() << endl;
	CDat ra = this->m_cpu->getRegister(m_ra);
	CDatd reg = ra - this->m_imm;
	this->m_cpu->updateFlags(reg, ra, (~this->m_imm)+1);
}

std::string Ccmpi::toString()
{
	stringstream op;
	op << this->getName();

	if(m_d) op << 'D';
	if(m_s) op << 'S';
	if(m_c) op << 'C';

	op << this->getConditionFlag() << " r" << m_ra << ", " << m_imm;

	return op.str();
}
